'''
Riffle- "A segment of stream where the flow is shallower and more turbulent."

Riffle is concerned with abstracting persistent TCP/TLS connections. Exhaustively, 
riffle is a deferred RPC protocol over TLS based on twisted.pb.PerspectiveBroker that uses Avatars 
automagically created from self-signed SSL keys. 

Conceptually, there are three important parts:
    - interacting with remote clients through RPC 
    - managing persistent connections
    - connecting or listening for Riffle connections

More to follow. 
'''

from twisted.python.filepath import FilePath
from twisted.internet.endpoints import SSL4ServerEndpoint
from twisted.internet.task import react
from twisted.internet import reactor
from twisted.internet import defer
from twisted.spread import pb
from twisted.cred import portal as twistedPortal
from twisted.internet.endpoints import SSL4ClientEndpoint
from twisted.internet.ssl import PrivateCertificate, Certificate, optionsForClientTLS

from zope.interface import implements
from pubsub import pub


from pdtools.lib.output import out


DEFAULT_PORT = 8016


############################################################
# Portal and Utility methods
############################################################

class Portal(twistedPortal.Portal):

    '''
    This is a portal for bi-directional communication between two parites. Both sides
    must have set up a portal in order to initiate communication. 

    Portals assign incoming connections to Realm objects based on the name assigned
    to the remote connection. Realms create avatars and return them across the wire. 

    Do not instantiate more than one portal-- don't instantiate it at all, actually.
    When serving or connecting, the global portal object in this class holds *all* 
    active connections.

    There are three instances where you may need to interact with the portal
        - initialization: set matchers and realms
        - access: find the connection with the given name or type
        - callback: assignable callbacks called when a new connection is made
            (Under consideration)
    '''

    def __init__(self):
        self.realms = {}
        self.host, self.port = 'localhost', DEFAULT_PORT

        self.certCa = None
        self.keyPrivate = None
        self.keyPublic = None

    def open(self, port=None, cert=None):
        '''
        Listen for connections on the given port. 
        '''
        port = port if port else self.port
        cert = cert if cert else self.certCa

        ca = Certificate.loadPEM(cert)
        myCertificate = PrivateCertificate.loadPEM(cert)

        SSL4ServerEndpoint(reactor, port, myCertificate.options(ca)).listen(RiffleServerFactory(self))

    @defer.inlineCallbacks
    def connect(self, host=None, port=None, cert=None, key=None):
        '''
        Connect to another portal.
        '''

        host = host if host else self.host
        port = port if port else self.port
        cert = cert if cert else self.certCa
        key = key if key else self.keyPrivate  # ???

        # the first term is the name the server is using in the cert (for now)
        ctx = optionsForClientTLS(u"pds.production", Certificate.loadPEM(cert), PrivateCertificate.loadPEM(key))

        factory = RiffleClientFactory()
        SSL4ClientEndpoint(reactor, host, port, ctx,).connect(factory)

        avatar = yield factory.login(self)

        defer.returnValue(Levy(avatar))

    def close(self):
        '''
        Close all connections in all realms.
        '''
        out.info("Portal closing all connections")
        for k, v in self.realms.iteritems():
            for c in v.connections:
                c.destroy()

    def addRealm(self, matcher, realm):
        '''
        Add a realm to this portal with its corresponding matcher. This can be done after
        the portal has been opened.

        :param realm: the realm object that will handle the incoming connections
        :type realm: riffle.Realm
        :param matcher: an re matcher object that tests presented domain names for incoming clients
        :type matcher: re.matcher
        '''
        self.realms[matcher] = realm

    def findRealm(self, credential):
        '''
        Find the appropriate realm for the given credential. Matches are found using
        re filters. 
        '''

        for k, v in self.realms.iteritems():
            if k.match(credential):
                return v

        raise KeyError("No matcher was found to handle ", credential)

    def getRealm(self, matcher):
        '''
        Returns a realm based on a broad query for realm types.
        '''
        return self.realms[matcher]

    def login(self, credentials, mind):
        # print 'Login request from ' + credentials
        target = self.findRealm(credentials)
        return target.requestAvatar(credentials, mind)

    def partialLogin(self, credentials):
        '''
        Clients send servers their representations before the server gives the client 
        its 'mind' object. This call returns the correct representation for the server
        *without* the mind object. Does not add the avatar to a realm. 
        '''
        target = self.findRealm(credentials)
        return target.requestPartialAvatar(credentials)

    def connectionForName(self, credentials):
        '''
        Find the connection that has the given credentials.

        :param credentials: object to query for
        :type credentials: str.
        :return: the connection avatar or None
        '''

        # print 'Looking for connection with name:', credentials

        r = self.findRealm(credentials)
        # print r
        # print 'Connections:', len(r.connections)

        for c in r.connections:
            # print c

            if c.name == credentials:
                return c

        return None

portal = Portal()


def dumpRealms(portal=portal):
    s = 'Dumping all connections\n'

    for k, v in portal.realms.iteritems():
        s += '\tRealm: %s\n' % v.avatar.__name__

        for c in v.connections:
            s += '\t\t%s\n' % c.name

    print s


############################################################
# Realms, Avatars, and Levies
############################################################

class Realm:

    '''
    Wraps a type of avatar and all connections for that avatar type.

    Broadcasts connection changes using pubsub. To be notfied of new connections:
        pubsub.pub.subscribe(con, '[avatarName]Connected')
        pubsub.pub.unsubscribe(dis, '[avatarName]Disconnected')

    where 'dis' and 'con' are method references. Both take keyword arguments 'type'
    and 'realm'.
    '''
    implements(twistedPortal.IRealm)

    def __init__(self, avatar):
        self.avatar = avatar
        self.connections = set()

    @defer.inlineCallbacks
    def requestAvatar(self, avatarID, mind):
        '''
        Returns an instance of the appropriate avatar. Asks the avatar to perform 
        any needed initialization (which should be a deferred)
        '''
        avatar = self.avatar(avatarID, self)

        # print 'Attaching %s to %s' % (mind, avatar)
        avatar.attached(mind)

        # print avatar.remote

        self.connections.add(avatar)
        out.info('Connected: ' + avatar.name)
        pub.sendMessage('%sConnected' % self.avatar.__name__, avatar=avatar, realm=self)

        # yield mind.callRemote('echo', '')

        # deferred since initialization may require database operations
        # yield avatar.initialize()
        yield 1

        # move detached from the avatar to here?
        defer.returnValue((avatar, lambda a=avatar: a.detached(mind)))

    def attach(self, avatar, mind):
        '''
        Temp method. Performs the other half of the partial avatar method.

        Leaving the requestAvatar method is most likely a bad, bad idea.
        The more riffle diverges from PB the greater chance for bugs across the board.
        '''
        avatar.attached(mind)

        self.connections.add(avatar)
        out.info('Connected: ' + avatar.name)

    def requestPartialAvatar(self, avatarID):
        return self.avatar(avatarID, self)

    def requestPartialAvatar(self, avatarID):
        return self.avatar(avatarID, self)

    def connectionClosed(self, avatar):
        out.info('Disconnected: ' + str(avatar.name))
        pub.sendMessage('%sDisconnected' % self.avatar.__name__, avatar=avatar, realm=self)
        self.connections.remove(avatar)


class Levy(object):

    ''' Wraps a remote object reference to allow getattr magic '''

    def __init__(self, remote):
        self.remote = remote

    def __getattr__(self, item):
        def wrap(*args):
            return self.remote.callRemote(item, *args).addCallbacks(self.printValue, self.printError)

        return wrap

    # def __repr__(self):
    #     return 'Levy wrapping:\n\t' + repr(self.remote)

    def printValue(self, value):
        # print repr(value)
        return value

    def printError(self, error):
        out.warn('Default riffle error trap: ' + str(error))


class RifflePerspective(pb.Avatar):

    def __init__(self, name, realm):
        self.name = name
        self.realm = realm
        self.remote = None

    @defer.inlineCallbacks
    def initialize(self):
        '''
        An initialization method that may hit the database or perform other model
        related tasks needed to initialize this avatar. This method is meant to be subclassed. 

        NEVER make riffle calls here. There is no guarantee
        the other end of the connection is finished loading by the time this end of the connection
        is. Register for portal callbacks instead: this is a model and init method.
        '''
        yield

    def destroy(self):
        '''
        Dealloc and destroy. This is most likely called for Bad Reasons, 
        but for now its a catch all for all dealloc. 

        TODO: drop the connection (from twisted's perspective)
        '''
        pass

    @defer.inlineCallbacks
    def perspective_handshake(self):
        '''
        Utility method. Called when both ends of the connection have come online. 
        Only purpose is to call the initialize method.

        Note: malicious endpoints can easily call this repeatedly. Add a check 
        to ensure init is only called once. 
        '''
        # return self.initialize()
        yield self.initialize()

    def attached(self, mind):
        self.remote = Levy(mind)

    def detached(self, mind):
        self.destroy()
        self.remote = None
        self.realm.connectionClosed(self)

    def perspective_echo(self, arg):
        print 'Echo from %s: "%s"' % (self.name, arg)
        return arg


class RiffleReferencable(pb.Referenceable):

    def __init__(self, name, realm):
        self.name = name

    def connected(self, perspective):
        ''' Called when a remote user gains access to this object. Must save a reference '''
        self.perspective = perspective


class RiffleViewable(pb.Viewable):

    def view_doFoo(self, perspective, arg1, arg2):
        print 'Do Foo!', perspective, arg1, arg2
        return 'Done'


############################################################
# Perspective Broker Monkey Patches
############################################################

class _RiffleBroker(pb.Broker):

    def connectionLost(self, reason):
        '''
        This triggers many times, but the reason its here now is to 
        detect invalid certs. Without the print statement, the client isnt warned. This is a 
        todo item. 
        '''
        print 'Connection Lost!', reason
        return pb.Broker.connectionLost(self, reason)


class RiffleClientFactory(pb.PBClientFactory):

    protocol = _RiffleBroker

    @defer.inlineCallbacks
    def login(self, portal):
        # Have to add connection to the portal
        self.portal = portal

        print 'Attempting to login!'
        # Returns a _RifflePortalWrapper remote reference
        root = yield self.getRootObject()
        print 'Got the root object!'

        # Extract the name from credentials
        peerCertificate = Certificate.peerFromTransport(self._broker.transport)
        pdid = peerCertificate.getSubject().commonName.decode('utf-8')

        # Returns the server's avatar based on the client's interpretation
        # 'other' is the other end of the connection.
        other = self.portal.partialLogin(pdid)

        # Convert to remote referentiable for transmission
        referencibleOther = pb.AsReferenceable(other, "perspective")

        # Here is the problem. The login call triggers the init call, and that
        # starts the chain of riffle callbacks, but this object won't be back in time by then!

        # Avatar is the remotely described API object
        avatar = yield root.callRemote('login', referencibleOther)
        other.remote = Levy(avatar)

        # print 'Return Avatar: ' + str(avatar)
        # print 'Other Remote: ' + str(other)
        # print 'Not returned avatar: ' + str(a)
        # print ': ' + str(a.remote)
        # print 'Returned Mind: ' + str(other)

        # Signal we're done cleaning up
        # print 'Calling Handshake'

        # other.remote.callRemote('echo', 'boo')

        # This is absolutely not needed. The point of this method is to registers
        # the new connections with the portal, but we already have all the pieces.
        # a, b = yield self.portal.login(pdid, avatar)

        # tttteeeemmmpppppoooorrraaaarrryyy
        realm = self.portal.findRealm(pdid)
        realm.attach(other, avatar)
        # a, b = self.portal.

        # print 'Realm Object: ' + str(a)

        # Bit hacky, but should be cleaned up in a refactor
        # a.remote = avatar

        # other.remote = avatar

        avatar.callRemote('handshake')
        other.perspective_handshake()

        defer.returnValue(avatar)


class RiffleServerFactory(pb.PBServerFactory):

    protocol = _RiffleBroker

    def __init__(self, portal):
        pb.PBServerFactory.__init__(self, portal)
        # print 'Server factory started'
        self.root = _RifflePortalRoot(portal)


class _RifflePortalRoot(pb._PortalRoot):

    def rootObject(self, broker):
        # print 'Returning the root object!'
        return _RifflePortalWrapper(self.portal, broker)


class _RifflePortalWrapper(pb._PortalWrapper):

    @defer.inlineCallbacks
    def remote_login(self, client):
        # print 'Remote login!'
        peerCertificate = Certificate.peerFromTransport(self.broker.transport)
        pdid = peerCertificate.getSubject().commonName.decode('utf-8')

        avatar, logout = yield self.portal.login(pdid, client)
        avatar = pb.AsReferenceable(avatar, "perspective")

        # Formerly in _cbLogin, moved here to make the deferred chain cleaner
        puid = avatar.processUniqueID()

        # only call logout once, whether the connection is dropped (disconnect)
        # or a logout occurs (cleanup), and be careful to drop the reference to
        # it in either case
        logout = [logout]

        def maybeLogout():
            if not logout:
                return
            fn = logout[0]
            del logout[0]
            fn()

        self.broker._localCleanup[puid] = maybeLogout
        self.broker.notifyOnDisconnect(maybeLogout)

        defer.returnValue(avatar)

    def __init__(self, portal, broker):
        # print 'Root object initializing'
        self.portal = portal
        self.broker = broker

        self.broker.notifyOnDisconnect(disc)
        self.broker.notifyOnFail(fail)
        self.broker.notifyOnConnect(connect)

    def remote_loginAnonymous(self, mind):
        # print 'Remote login!'
        return None

# Do we need these? Who knows. Keeping them around for know, although
# these are a full level of abstraction below where we want to be, so be careful using them


def disc():
    # print 'CB: Disconnected'
    pass


def fail():
    # print 'CB: Failed'
    pass


def connect():
    # print 'CB: Connected'
    pass



'''
All kinds of encryption and security concerns
'''

import bcrypt


def hashPassword(password):
    salt = bcrypt.gensalt()
    hashed = bcrypt.hashpw(str(password), salt)
    return hashed


def checkPassword(raw, hashed):
    ''' Returns true if a hashed password matches a raw one '''
    return bcrypt.hashpw(str(raw), str(hashed)) == hashed



'''
Communication with routers
'''

import requests
import json
import urllib
import os
import yaml
from os.path import expanduser

from twisted.internet import defer

from pdtools.coms import general
from pdtools.coms.client import RpcClient
from pdtools.lib.store import store
from pdtools.lib import pdutils, riffle
from pdtools.lib.output import out


###############################################################################
# Riffle and Non-riffle implementations
###############################################################################

@general.defaultCallbacks
@defer.inlineCallbacks
def provisionRouter(r, name, host, port):

    # Bug: if the user sets the name of the router as their username, this will
    # fail badly
    target = [x for x in store.getConfig('routers') if name in x['_id']]

    if len(target) == 0:
        print 'Router with name ' + name + ' not found.'
        yield 1
        defer.returnValue(None)

    target = target[0]

    pkey = store.getKey(target['_id'] + '.client.pem')
    cacert = store.getKey('ca.pem')

    client = RpcClient(host, port, 'internal')
    ret = yield client.provision(target['_id'], pkey, cacert)


###############################################################################
# Chute Operations
###############################################################################

def installChute(host, port, config):
    '''
    Take a local config yaml file and launch chute of given host with pdfcd running on specified port.
    '''

    # Read local yaml into a config_json to send via post request
    try:
        with open(config, 'r') as stream:
            config_json = yaml.load(stream)
            config_json['date'] = config_json['date'].isoformat()
    except IOError as e:
        print 'Error: ' + config + ' not found.'
        return

    # Verify the config provided in some way.
    cfg_verf = pdutils.check(config_json, dict, {'dockerfile': dict, 'name': str, 'owner': str})
    if cfg_verf:
        print 'ERROR: ' + cfg_verf
        return

    # Only allowed one way to provide a dockerfile order is local, remote, inline
    if 'local' in config_json['dockerfile']:
        with open(config_json['dockerfile']['local'], 'r') as stream:
            config_json['dockerfile'] = stream.read()
    elif 'remote' in config_json['dockerfile']:
        print 'Remote Dockerfile not supported yet.'
        return
    elif 'inline' in config_json['dockerfile']:
        config_json['dockerfile'] = config_json['dockerfile']['inline']
    else:
        print 'ERROR: No Dockerfile specified in config file.'
        return

    print 'Installing chute...\n'
    params = {'config': config_json}
    r = requests.post('http://' + host + ':' + str(port) + '/v1/chute/create', data=json.dumps(params), stream=True)
    for line in r.iter_lines():
        if line:
            try:
                line = json.loads(line)
                if line.get('success'):
                    print line.get('message')
                else:
                    print 'ERROR: Failed to install chute.(' + urllib.unquote(str(line.get('message'))) + ')'
            except Exception as e:
                print line


def deleteChute(host, port, name):
    '''
    Remove chute with given name from host with pdfcd running on specified port.
    '''

    print 'Removing chute...\n'

    params = {'name': name}
    r = requests.post('http://' + host + ':' + str(port) + '/v1/chute/delete', data=json.dumps(params), stream=True)

    for line in r.iter_lines():
        if line:
            try:
                line = json.loads(line)
                if line.get('success'):
                    print line.get('message')
                else:
                    print 'ERROR: Failed to delete chute.(' + urllib.unquote(str(line.get('message'))) + ')'
            except Exception as e:
                print line


def stopChute(host, port, name):
    '''
    Stop chute with given name from host with pdfcd running on specified port.
    '''

    print 'Stopping chute...\n'

    params = {'name': name}
    r = requests.post('http://' + host + ':' + str(port) + '/v1/chute/stop', data=json.dumps(params), stream=True)

    for line in r.iter_lines():
        if line:
            try:
                line = json.loads(line)
                if line.get('success'):
                    print line.get('message')
                else:
                    print 'ERROR: Failed to stop chute.(' + urllib.unquote(str(line.get('message'))) + ')'
            except Exception as e:
                print line


def startChute(host, port, name):
    '''
    Start chute with given name from host with pdfcd running on specified port.
    '''

    print 'Starting chute...\n'

    params = {'name': name}
    r = requests.post('http://' + host + ':' + str(port) + '/v1/chute/start', data=json.dumps(params), stream=True)

    for line in r.iter_lines():
        if line:
            try:
                line = json.loads(line)
                if line.get('success'):
                    print line.get('message')
                else:
                    print 'ERROR: Failed to start chute.(' + urllib.unquote(str(line.get('message'))) + ')'
            except Exception as e:
                print line



'''
Generic communication tasks. 

All remote calls are inline deferred callbacks. This allows them to be
easily chained together and fired asynchronously. 

Since calls are started with task.react, each method must take a reactor as its 
first parameter. 

Adding the defaultCallbacks decorator prints the result of the operation on 
success or failure. All failed calls return an exception which should be 
suitable for printing. 
'''

import re

from twisted.internet import defer

from pdtools.coms.client import RpcClient
from pdtools.lib.output import out
from pdtools.lib import pdutils, store, riffle, names


###############################################################################
# Boilerplate Callbacks
###############################################################################

def failureCallbacks(f):
    def w(*args, **kwargs):
        return f(*args, **kwargs).addErrback(printFailure)

    return w


def defaultCallbacks(f):
    ''' 
    Wrap the API call in default handlers which print results. Optional 
    params are used to turn off specific handlers
    '''
    def w(*args, **kwargs):
        return f(*args, **kwargs).addCallbacks(printSuccess, printFailure)

    return w


def printSuccess(r):
    if isinstance(r, list):
        for x in r:
            print x
    else:
        print r
    return r


def printFailure(r):
    print r
    return r


###############################################################################
# Generic Functionality
###############################################################################

@defaultCallbacks
@defer.inlineCallbacks
def echo(reactor, host, port):
    avatar = yield riffle.portal.connect(host=host, port=port)
    response = yield avatar.echo("Hello!")
    defer.returnValue(response)


@defer.inlineCallbacks
def test(r):
    avatar = yield riffle.portal.connect()
    response = yield avatar.test()

    contents = yield response.callRemote('stuff')
    print contents

    defer.returnValue(response)



'''
Communication with the server
'''

import getpass

from twisted.internet import defer

from pdtools.coms import general
from pdtools.coms.client import RpcClient
from pdtools.lib.store import store
from pdtools.lib import riffle, names
from pdtools.lib.output import out


###############################################################################
# Riffle Implementation
###############################################################################
class ServerPerspective(riffle.RifflePerspective):

    def perspective_logs(self, logs):
        ''' New logs coming in from the server '''
        for x in logs:
            print out.messageToString(x)


@defer.inlineCallbacks
def list(r):
    ''' Return the resources this us4er owns. '''

    avatar = yield riffle.portal.connect()
    ret = yield avatar.list()

    # I don't think this is necesary anymore, but keeping it here for now
    # Have to hit the server every call anyway, so whats the point of saving these
    # simple lists? User will not be able to do anything offline anyway.
    store.saveConfig('chutes', ret['chutes'])
    store.saveConfig('routers', ret['routers'])
    store.saveConfig('instances', ret['instances'])

    printOwned()

    defer.returnValue(ret)


# Merge this with provision!
@general.failureCallbacks
@defer.inlineCallbacks
def createRouter(r, name):
    '''
    Create a new router on the server-- do not provision it yet. 

    Like so many other things, this is a temporary method.
    '''

    avatar = yield riffle.portal.connect()
    ret = yield avatar.provisionRouter(name)

    # Save that router's keys
    store.saveKey(ret['keys'], ret['_id'] + '.client.pem')

    ret = yield avatar.list()

    store.saveConfig('chutes', ret['chutes'])
    store.saveConfig('routers', ret['routers'])
    store.saveConfig('instances', ret['instances'])

    print 'New router successfully created'
    printOwned()

    defer.returnValue("Done")


@defer.inlineCallbacks
def logs(r, pdid):
    '''
    Query the server for all logs that the given pdid has access to. Must be a fully qualified name.

    NOTE: this method is in progress. For now, just pass the name of one of your routers.
    '''

    # Let the validation occur serverside (or skip it for now)
    pdid = store.getConfig('pdid') + '.' + pdid

    avatar = yield riffle.portal.connect()
    ret = yield avatar.logs(pdid)


@general.defaultCallbacks
@defer.inlineCallbacks
def test(r):
    ''' Should be able to receive a model when prompted '''
    avatar = yield riffle.portal.connect(host='localhost')
    ret = yield avatar.test()

    # print ret.__dict__
    name = yield ret.callRemote('data')
    print 'Result from call: ' + name

###############################################################################
# Authentication
###############################################################################


def authCallbacks(f):
    def w(*args, **kwargs):
        return f(*args, **kwargs).addCallbacks(authSuccess, general.printFailure)

    return w


def authSuccess(r):
    store.saveConfig('pdid', r['_id'])
    store.saveKey(r['keys'], 'client.pem')
    store.saveKey(r['ca'], 'ca.pem')

    print 'You have been successfully logged in.'


@authCallbacks
@defer.inlineCallbacks
def login(reactor, host, port):
    name, password = None, None

    name = raw_input("Username: ")
    password = getpass.getpass()

    client = RpcClient(host, port, '')
    ret = yield client.login(name, password)
    defer.returnValue(ret)


@authCallbacks
@defer.inlineCallbacks
def register(reactor, host, port):
    name, email, password = raw_input("Username: "), raw_input("Email: "), getpass.getpass()

    client = RpcClient(host, port, '')
    ret = yield client.register(name, email, password)
    defer.returnValue(ret)


###############################################################################
# Utils
###############################################################################

def printOwned():
    chutes = store.getConfig('chutes')
    routers = store.getConfig('routers')

    def pprint(d):
        print '\t' + d['_id']

    print 'routers'
    [pprint(x) for x in routers]

    print 'chutes'
    [pprint(x) for x in chutes]



'''
Client-side tools for communicating with a server using RPC
'''

from twisted.web.xmlrpc import Proxy


class RpcClient:

    '''
    Remote client RPC wrapper. Translates seemingly local calls
    into remote RPC calls to the given server. 

    Usage: 
        client = RpcClient("https://something:70234")
        call = client.nameOfRemoteCall(arg1 arg2)
        call.addCallbacks(aCallback)


    API calls always return deferreds. 
    '''

    def __init__(self, host, port, path):
        self.proxy = Proxy('http://%s:%s/%s/' % (host, port, path), allowNone=True)

    def __getattr__(self, item):
        def wrap(*args):
            return self.proxy.callRemote(item, *args)

        return wrap



###################################################################
# Copyright 2013-2015 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

"""
lib.utils.output.
Helper for formatting output from Paradrop.
"""

import time
import json
import urllib

timeflt = lambda: time.time()
timeint = lambda: int(time.time())
timestr = lambda x=None: time.asctime(time.localtime(x)) if x else time.asctime()

# Convert localtime to gmtime! Serious!


def timedur(x):
    """
        Print consistent string format of seconds passed.
        Example: 300 = '5 mins'
        Example: 86400 = '1 day'
        Example: 86705 = '1 day, 5 mins, 5 sec'
    """
    divs = [('days', 86400), ('hours', 3600), ('mins', 60)]
    x = float(x)
    res = []
    for lbl, sec in divs:
        if(x >= sec):
            rm, x = divmod(x, float(sec))
            # If exactly 1, remove plural of label
            if(rm == 1.0):
                res.append((lbl[:-1], int(rm)))
            else:
                res.append((lbl, int(rm)))

    # anything left over is seconds
    x = int(x)
    if(x == 1):
        res.append(("second", x))
    elif(x == 0):
        pass
    else:
        res.append(("seconds", x))

    return ", ".join(["%d %s" % (x[1], x[0]) for x in res])


def convertUnicode(elem):
    """Converts all unicode strings back into UTF-8 (str) so everything works.
        Call this function like:
            json.loads(s, object_hook=convertUnicode)"""
    if isinstance(elem, dict):
        return {convertUnicode(key): convertUnicode(value) for key, value in elem.iteritems()}
    elif isinstance(elem, list):
        return [convertUnicode(element) for element in elem]
    elif isinstance(elem, unicode):
        return elem.encode('utf-8')
    # DFW: Not sure if this has to be here, but deal with possible "null" MySQL strings
    elif(elem == 'null'):
        return None
    else:
        return elem


def urlEncodeMe(elem, safe=' '):
    """
        Converts any values that would cause JSON parsing to fail into URL percent encoding equivalents.
        This function can be used for any valid JSON type including str, dict, list.
        Returns:
            Same element properly encoded.
    """
    # What type am I?
    if isinstance(elem, dict):
        return {urlEncodeMe(key, safe): urlEncodeMe(value, safe) for key, value in elem.iteritems()}
    elif isinstance(elem, list):
        return [urlEncodeMe(element, safe) for element in elem]
    elif isinstance(elem, str):
        # Leave spaces alone, they are save to travel for JSON parsing
        return urllib.quote(elem, safe)
    else:
        return elem


def urlDecodeMe(elem):
    """
        Converts any values that would cause JSON parsing to fail into URL percent encoding equivalents.
        This function can be used for any valid JSON type including str, dict, list.
        Returns:
            Same element properly decoded.
    """
    # What type am I?
    if isinstance(elem, dict):
        return {urlDecodeMe(key): urlDecodeMe(value) for key, value in elem.iteritems()}
    elif isinstance(elem, list):
        return [urlDecodeMe(element) for element in elem]
    elif isinstance(elem, str):
        # Leave spaces alone, they are save to travel for JSON parsing
        return urllib.unquote(elem)
    else:
        return elem


def jsonPretty(j):
    """
        Returns a string of a JSON object in 'pretty print' format fully indented, and sorted.
    """
    return json.dumps(j, sort_keys=True, indent=4, separators=(',', ': '))

'''
These two methods are outrageously slow.

Example with 50k records in var 'logs':
    from pdtools.lib import pdutils
    import json

    with pdutils.Timer() as t:
        [pdutils.str2json(x) for x in logs]

    with pdutils.Timer() as t:
        [json.loads(x) for x in logs]

    >> time: 12814.471006 ms
    >> time: 0.007868 ms

Vanilla, sanitized benches do not show the same performance issues. Not sure why, 
it might have to do with the content of the dictionaries or the complexity of the underlying
data? They're still slower, but only by one or two magnitudes.
'''
def json2str(j, safe=' '):
    """
        Properly converts and encodes all data related to the JSON object into a string format
        that can be transmitted through a network and stored properly in a database.
        Arguments:
            @j    : json to be converted
            @safe : optional, string of chars to pass to urlEncodeMe that are declared safe (don't encode)
    """
    return json.dumps(urlEncodeMe(j, safe), separators=(',', ':'))


def str2json(s):
    t = json.loads(s, object_hook=convertUnicode)
    # If t is a list, object_hook was never called (by design of json.loads)
    # deal with that situation here
    if(isinstance(t, list)):
        t = [convertUnicode(i) for i in t]
    # Make sure to still decode any strings
    return urlDecodeMe(t)


class dict2obj(object):

    def __init__(self, aDict=None, **kwargs):
        if(aDict is not None):
            aDict.update(kwargs)
        else:
            self.__dict__.update(kwargs)


def check(pkt, pktType, keyMatches=None, **valMatches):
    """This function takes an object that was expected to come from a packet (after it has been JSONized)
        and compares it against the arg requirements so you don't have to have 10 if() statements to look for keys in a dict, etc..

        Args:
            @pkt             : object to look at
            @pktType         : object type expected (dict, list, etc..)
            @keyMatches      : a list of minimum keys found in parent level of dict, expected to be an array
            @valMatches      : a dict of key:value pairs expected to be found in the parent level of dict
                              the value can be data (like 5) OR a type (like this value must be a @list@).
        Returns:
            None if everything matches, otherwise it returns a string as to why it failed."""
    # First check that the pkt type is equal to the input type
    if(type(pkt) is not pktType):
        return 'expected %s' % str(pktType)

    if(keyMatches):
        # Convert the keys to a set
        keyMatches = set(keyMatches)
        # The keyMatches is expected to be an array of the minimum keys we want to see in the pkt if the type is dict
        if(type(pkt) is dict):
            if(not keyMatches.issubset(pkt.keys())):
                return 'missing, "%s"' % ', '.join(list(keyMatches - set(pkt.keys())))
        else:
            return None

    # Finally for anything in the valMatches find those values
    if(valMatches):
        # Pull out the dict object from the "valMatches" key
        if('valMatches' in valMatches.keys()):
            matchObj = valMatches['valMatches']
        else:
            matchObj = valMatches

        for k, v in matchObj.iteritems():
            # Check for the key
            if(k not in pkt.keys()):
                return 'key missing "%s"' % k

            # See how we should be comparing it:
            if(type(v) is type):
                if(type(pkt[k]) is not v):
                    return 'key "%s", bad value type, "%s", expected "%s"' % (k, type(pkt[k]), v)

            else:
                # If key exists check value
                if(v != pkt[k]):
                    return 'key "%s", bad value data, "%s", expected "%s"' % (k, pkt[k], v)

    return None


def explode(pkt, *args):
    """This function takes a dict object and explodes it into the tuple requested.

        It returns None for any value it doesn't find.

        The only error it throws is if args is not defined.

        Example:
            pkt = {'a':0, 'b':1}
            0, 1, None = pdcomm.explode(pkt, 'a', 'b', 'c')
    """
    if(not args):
        raise PDError('EXPLODE', 'args must be provided')

    # If there is an error make sure to return a tuple of the proper length
    if(not isinstance(pkt, dict)):
        return tuple([None] * len(args))

    # Now just step through the args and pop off everything from the packet
    # If a key is missing, the pkt.get(a, None) returns None rather than raising an Exception
    return tuple([pkt.get(a, None) for a in args])


class Timer(object):
    '''
    A timer object for simple benchmarking. 

    Usage:
        with Timer(key='Name of this test') as t:
            do.someCode(thatTakes=aWhile)

    Once the code finishes executing the time is output. 
    '''
    def __init__(self, key="", verbose=True):
        self.verbose = verbose
        self.key = key

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args):
        self.end = time.time()
        self.secs = self.end - self.start
        self.msecs = self.secs * 1000  # millisecs
        if self.verbose:
            print self.key + ' elapsed time: %f ms' % self.msecs


'''
Exceptions and their subclasses

TODO: Distill these down and make a heirarchy.
'''


class PdServerException(Exception): pass


class InteralException(Exception): pass


class AuthenticationError(PdServerException): pass


class InvalidCredentials(PdServerException): pass


class PdidError(PdServerException): pass


class PdidExclusionError(PdServerException): pass



'''
Entry point for client tools.

This module has two logical sections: extracting and validating arguments, and 
sending program flow into the correct function.

Structurally there are three sections. In order:
    - Literal docstrings users see
    - Subcommand handlers
    - Entry point

The main method calls docopt, which parses the first positional command given.
If that command matches a sub-command the correct handler is invoked. 

Docopt will not allow execution to continue if the args matcher fails (in other words:
dont worry about failure conditions.) Check documentation for more information.
'''

from pkg_resources import get_distribution

from docopt import docopt
from twisted.internet import task
from twisted.internet import reactor

from pdtools.lib import output, riffle, names
from pdtools.coms import routers, general, server
from pdtools.lib.store import store


SERVER_HOST = 'paradrop.io'
# SERVER_HOST = 'localhost'
SERVER_PORT = 8015  # this is the vanilla server port, not the riffle one


rootDoc = """
usage: paradrop [options] <command> [<args>...]
        
options:
   -h, --help
   --version
   -v, --verbose    Show verbose internal output       
   
commands:
    router     Manage routers
    chute      Manage chutes
    
    list       List and search for resources you own
    logs       Query logs
    
    login      
    register   Not implemented.
    logout

See 'paradrop <command> -h' for more information on a specific command.    
"""

routerDoc = """
usage: 
    paradrop router create <name> 
    paradrop router provision <name> <host> <port>

commands: 
    create      Create a new router with the given name
    provision   Assign a newly created router identity to a physical instance
"""

chuteDoc = """
usage:
    paradrop chute install <host> <port> <path-to-config>
    paradrop chute delete <host> <port> <name>
    paradrop chute start <host> <port> <name>
    paradrop chute stop <host> <port> <name>

commands: 
    start       Start the installed chute with the given name
    stop        Stop the installed chute with the given name
    install     Install a chute on the given router
    delete      Delete the installed chute on the given router
"""

listDoc = """
usage: paradrop list

Lists all owned resources.
"""

logsDoc = """
usage: paradrop logs <name>

Displays the logs for the provided resource. The resource, commonly a router, 
must be online.
"""


def routerMenu():
    args = docopt(routerDoc, options_first=False)

    checkLoggedIn()

    if args['provision']:
        task.react(routers.provisionRouter, (args['<name>'], args['<host>'], args['<port>']))

    elif args['create']:
        task.react(server.createRouter, (args['<name>'],))

    else:
        print routerDoc


def chuteMenu():
    args = docopt(chuteDoc, options_first=False)

    checkLoggedIn()

    if args['install']:
        return routers.installChute(args['<host>'], args['<port>'], args['<path-to-config>'])

    if args['delete']:
        return routers.deleteChute(args['<host>'], args['<port>'], args['<name>'])

    if args['start']:
        return routers.startChute(args['<host>'], args['<port>'], args['<name>'])

    if args['stop']:
        return routers.stopChute(args['<host>'], args['<port>'], args['<name>'])

    print routerDoc


def listMenu():
    args = docopt(listDoc)

    checkLoggedIn()

    # currently only one thing to do, but this will get more hefty
    task.react(server.list)


def logsMenu():
    args = docopt(logsDoc)
    checkLoggedIn()

    reactor.callLater(.1, server.logs, None, args['<name>'])
    reactor.run()
    exit(0)


###################################################################
# Utility and Initialization
###################################################################

def checkLoggedIn():
    # logged in calls
    if not store.loggedIn():
        print 'You must login first.'
        exit(1)


def setup(displayToConsole=False, logLevel=0):
    '''
    Boilerplate setup. Start the logger, give riffle crypto keys, and 
    initialize riffle's portal by creating name to realm assignments
    '''
    # For now, don't grab STDIO and don't write random log noise to conosle
    output.out.startLogging(stealStdio=False, printToConsole=displayToConsole)

    # Initialize riffle with default values (can be overridden later)
    # NOTE: riffle serves on its own default port. This is a different port from the const above
    riffle.portal.certCa = store.getKey('ca.pem')
    riffle.portal.keyPrivate = store.getKey('client.pem')
    riffle.portal.host = SERVER_HOST

    # Register realms. See riffle documentation for Realm in pdtools.lib.riffle
    riffle.portal.addRealm(names.matchers[names.NameTypes.server], riffle.Realm(server.ServerPerspective))
    riffle.portal.addRealm(names.matchers[names.NameTypes.router], riffle.Realm(riffle.RifflePerspective))


def main():
    # present documentation, extract arguments
    args = docopt(rootDoc, version=get_distribution('pdtools').version, options_first=True, help=True)
    argv = [args['<command>']] + args['<args>']
    command = args['<command>']

    # Check for verbose flag. If set, turn on the serious logging.
    # TODO: set lower level log filters based on the number of '-v's passed in
    setup(displayToConsole=args['--verbose'])

    # Unpublished functions-- these here for testing purposes
    if command == 'echo':
        task.react(general.echo, (args['<args>'][0], int(args['<args>'][1]),))

    if command == 'test':
        task.react(server.test)

    if command == 'login':
        task.react(server.login, (SERVER_HOST, SERVER_PORT,))

    if command == 'register':
        task.react(server.register, (SERVER_HOST, SERVER_PORT,))

    # Check for a sub-command. If found, pass off execution to the appropriate sub-handler
    if command in 'router chute list logs'.split():
        return eval('%sMenu' % args['<command>'])()

    exit("%r is not a paradrop command. See 'paradrop -h'." % args['<command>'])

if __name__ == '__main__':
    main()



'''
Naming and permissions.

All objects are identified by a human-readable UUID, which is a domain-style
identifier. These are PDIDs.

All PDID's are prefaced with a top-level name. This currently serves no purpose
other than to make it obvious when a PDID appears, but could be used later
for alternative namespacing (development mode, 3rd party servers, etc).

There is exactly one special case when it comes to PDIDs in order to deal with
a conflict: specifying a chute "in the store." A user's top level namespace
is always followed by either a router name: pd.damouse.routerName OR a
chute namespace declaration: pd.damouse.chutes.chuteName. Queries for
pd.damouse.chuteName will be oppertunistically queried, but collisions must
ask the user to resolve the conflict.

Examples:
    User 'damouse':
        pd.damouse

    Chute 'netflix':
        pd.damouse.chutes.netflix

    Router 'aardvark':
        pd.damouse.aardvark

    Chute instance 'netflix':
        pd.damouse.aardvark.netflix


Full ids cannot have namespace collisions by definition. This has a few consequences:
    Chute names must be unique
    Owned router names must be unique

TODO:
    return constant values from idValid instead of strings
    Subdomaining should not require enumeration, fix.

Note that Groups, Organizations, and Chute Versions were previously pd.namespace objects, but
they do not offer enough flexibility-- these are now metatdata fields (but this could change)
'''

from pdtools.lib.exceptions import *

from enum import Enum
import re


###################################################
# Precompiled matchers for performance
###################################################

# A basic, valid name for anything
n = r'[a-zA-Z0-9]{3,32}'
basic = re.compile(r'pd\.%s' % n)

NameTypes = Enum('NameTypes', 'user, chute, router, instance, server')

matchers = {
    NameTypes.user: re.compile(r'^pd\.%s$' % n),
    NameTypes.chute: re.compile(r'^pd\.%s\.chutes\.%s$' % (n, n)),
    NameTypes.router: re.compile(r'^pd\.%s\.%s$' % (n, n)),
    NameTypes.instance: re.compile(r'^pd\.%s\.%s\.%s$' % (n, n, n)),
    NameTypes.server: re.compile(r'^pds.production$')
}


###################################################
# Formatting and Construction
###################################################

def idForUser(username):
    return 'pd.' + username


def idForChute(username, chuteName):
    return idForUser(username) + '.chutes.' + chuteName


def idForRouter(username, routerName):
    return idForUser(username) + '.' + routerName


def idForInstance(username, routerName, chuteName):
    return idForRouter(username, routerName) + '.' + chuteName


###################################################
# Validation
###################################################

def idValid(pdid):
    '''
    Checks any pdid for valid structure, but does not check that it is a
    valid reference (in other words it does not check the database)

    :returns: NameType specifying the type of this pdid
    '''

    if not basic.match(pdid):
        raise PdidError("Indeterminate PDID.")

    for k, v in matchers.iteritems():
        if v.match(pdid):
            return k

    # Might want to make this a little more specfic
    raise PdidError("Indeterminate PDID. Names must be between 3 and 32 characters.")



'''
I know this is confusing naming, this is a temp fix
'''

from setuptools import setup, find_packages

setup(
    name="pdtools",
    version="0.1.51",
    author="Paradrop Labs",
    description="Paradrop deployment and build tools",
    install_requires=[
        'bcrypt>=2.0.0',
        'cffi>=1.1.2',
        'colorama>=0.3.3',
        'docopt>=0.6.2',
        'enum>=0.4.4',
        'enum34>=1.0.0',
        'pyyaml>=3.11',
        'requests>=2.7.0',
        'service-identity>=14.0.0',
        'twisted>=14.2',
        'pypubsub>=3.3.0'
    ],

    packages=find_packages(),

    entry_points={
        'console_scripts': [
            'paradrop=pdtools.main:main',
        ],
    },
)
